import React, {useContext} from "react";
import {Instance} from "./Instance";
import {ColumnKey, Columns} from "./Column";
import {RowDefinition} from "./Row";
import {Promisable} from "@kernelui/core";

/**
 * Smartable data type.
 */
export type SmartableData<CK extends ColumnKey> = Promisable<(Promisable<RowDefinition<CK>>)[]>;

/**
 * Smartable properties.
 */
export interface SmartableProperties<CK extends ColumnKey>
{
	/**
	 * Table data.
	 */
	data: SmartableData<CK>;

	/**
	 * When defined, only show the columns in the array.
	 */
	shownColumns?: CK[];

	/**
	 * Default row element.
	 */
	rowElement?: React.ReactElement;

	/**
	 * Default column heading element.
	 */
	columnHeadingElement?: React.ReactElement;
}

/**
 * Smartable main object.
 */
export type Smartable<CK extends ColumnKey> = {
	/**
	 * Table component.
	 */
	Table: React.FunctionComponent<SmartableProperties<CK>>;
};

/**
 * Define a new Smartable.
 */
export function createSmartable<CK extends ColumnKey>({columns}: {
	/**
	 * Smartable table columns.
	 */
	columns: Columns<CK>;
}): Smartable<CK>
{
	/**
	 * Filter columns, given an array of shown columns.
	 * @param columns Full columns definition.
	 * @param shownColumns Shown columns list.
	 */
	const filterColumns = (columns: Columns<CK>, shownColumns: CK[]) => {
		// Initialize the filtered columns object.
		const filteredColumns = {...columns};

		// Build an associative object of shown columns.
		const shownColumnsMap = Object.fromEntries(shownColumns.map((shownColumn) => [shownColumn, true])) as Record<CK, boolean>;

		for (const key of Object.keys(columns) as CK[])
		{ // For each key, if it is not present in the shown columns
			if (!shownColumnsMap?.[key])
				// Delete all columns not present in the shown columns map.
				delete filteredColumns[key];
		}

		// Return filtered columns.
		return filteredColumns;
	};

	return {
		Table: (props: SmartableProperties<CK>) => {
			// Filter columns from the given property.
			const filteredColumns = props.shownColumns ? filterColumns(columns, props.shownColumns) : columns;

			return (
				<TableContext.Provider value={{
					columns: filteredColumns,
					...props,
				}}>
					<Instance columns={filteredColumns} {...props} />
				</TableContext.Provider>
			);
		},

	};
}

/**
 * Smartable table context data.
 */
export interface TableContextData<CK extends ColumnKey> extends SmartableProperties<CK>
{
	/**
	 * Current table columns.
	 */
	columns: Columns<CK>;
}

/**
 * Smartable table context.
 */
const TableContext = React.createContext<TableContextData<ColumnKey>>(undefined);

/**
 * Hook to get current Smartable table context.
 * @param smartable The Smartable main object.
 */
export function useTable<CK extends ColumnKey>(smartable?: Smartable<CK>): TableContextData<CK>
{
	return useContext(TableContext) as TableContextData<CK>;
}
